from typing import List

from collections import namedtuple

from matchup.structure.weighting.Tf import TF

Result = namedtuple("Result", "document, score")


class Solution:
    def __init__(self, results: List[Result]):
        self._results = results
        self._idf = None
        self._tf = None

    def __repr__(self):
        string = ""
        if self._results:
            for terms in self._results:
                doc = terms.document.split('\\')[-1]
                string += f"\n{doc} : {terms.score}"
        else:
            string += "No results found."
        return string

    def __contains__(self, item: Result):
        return item in self._results

    def expand(self, vocabulary):
        idf = vocabulary.idf
        self._idf = idf.take(reverse=True)
        self._tf = TF.take(reverse=True)

    def get_dict(self):
        result_list = []
        for terms in self._results:
            doc = terms.document.split('\\')[-1]
            result_list.append(Result(doc, terms.score))

        response_dict = {
            'solution': result_list,
            'stats': [
                {'idf': self._idf},
                {'tf': self._tf}
            ]
        }
        return response_dict



import unittest
import os

from matchup.presentation.Text import Term
from matchup.structure.Query import Query
from matchup.structure.Vocabulary import Vocabulary


class QueryTest(unittest.TestCase):
    def setUp(self):
        vocabulary = Vocabulary(settings_path=os.path.abspath("tests/static"),
                                processed_path=os.path.abspath("tests/static/files"))
        vocabulary.import_vocabulary()

        self._query = Query(vocabulary=vocabulary)

    def test_ask(self):
        answer = 'eu sou o marcos\ntop de linha'
        processed = [Term("marcos", "1-9"), Term("top", "2-0"), Term("linha", "2-7")]

        self._query.ask(answer=answer)
        self.assertEqual(self._query.search_input, processed)



from abc import ABC
from math import log
from collections import defaultdict


class TFFactory:
    @staticmethod
    def create_tf_by_str(value: str) -> "IDF":
        if value.lower() == 'binary':
            return Binary()
        elif value.lower() == 'termfrequency':
            return TermFrequency()
        elif value.lower() == 'lognormalization':
            return LogNormalization()


class TF(ABC):

    def __init__(self):
        self._tfs = defaultdict(float)

    def __repr__(self):
        string = ""
        for key in self._tfs:
            string += f"{key} : {self._tfs[key]}\n"
        return string

    def take(self, *,  value: int = 0, reverse: bool = True):
        if value != 0:
            return sorted(self._tfs.items(), key=lambda v: v[1], reverse=reverse)[:value]
        return sorted(self._tfs.items(), key=lambda v: v[1], reverse=reverse)

    def calculate(self, keyword: str, occurrence, vocabulary) -> float:
        """
           Generate TF based in TFType
        :return:float tf
        """
        ...

    @classmethod
    def _fij(cls, occurrence, vocabulary):
        if cls != LogNormalization:
            value = occurrence.frequency / (vocabulary.max_frequency_map[occurrence.doc()])
        else:
            value = occurrence.frequency
        return value

    def _persist(self, key, tf) -> float:
        self._tfs[key] = tf
        return tf


class Binary(TF):

    def __init__(self):
        super(Binary, self).__init__()

    def calculate(self, keyword: str, occurrence, vocabulary):
        """
            Model to calculate TF based in binary values (0,1)
        :return: float tf
        """
        fij = self._fij(occurrence, vocabulary)

        if fij > 0:
            return self._persist((keyword, occurrence.doc()), 1.0)
        else:
            return self._persist((keyword, occurrence.doc()), 0.0)


class TermFrequency(TF):

    def __init__(self):
        super(TermFrequency, self).__init__()

    def calculate(self, keyword: str, occurrence, vocabulary) -> float:
        """
            Model to calculate TF based in fij
        :return: float fij
        """
        fij = self._fij(occurrence, vocabulary)

        return self._persist((keyword, occurrence.doc()), float(fij))


class LogNormalization(TF):

    def __init__(self):
        super(LogNormalization, self).__init__()

    def calculate(self, keyword: str, occurrence, vocabulary) -> float:
        """
            Model to calculate TF based in : 1 + log fij
        :return: float fij
        """
        fij = self._fij(occurrence, vocabulary)
        tf = 1 + log(fij, 2)

        return self._persist((keyword, occurrence.doc()), tf)




"""
    Classic IR model. Probabilistic model are implemented here, based in Interface Model
"""
from collections import defaultdict
from typing import List, DefaultDict
from math import log

from matchup.models.Model import IterModel, Term, Vocabulary
from matchup.structure.Solution import Result


class Probabilistic(IterModel):

    # V : Conjunto solução
    _V = defaultdict(float)
    
    @staticmethod
    def run(query: List[Term], vocabulary: Vocabulary):
        """
            Principal method that represents IR probabilistic model.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return:
        """
        IterModel.initialize_occurrences(query, vocabulary)

        IterModel.initialize_pointers()

        # one iteration of probabilistic method
        rank = Probabilistic.__iter_rank(vocabulary)

        n_rank = Probabilistic.__take(rank, 5)

        if n_rank == Probabilistic._V:
            return rank
        else:
            Probabilistic._V = n_rank
            return Probabilistic.run(query, vocabulary)

    @staticmethod
    def __generate_term_scores(vocabulary: Vocabulary) -> DefaultDict[str, float]:
        """
            Generate scores of all terms in term occurrences based in Probabilistic model
        :param vocabulary: data structure representing vocabulary
        :return: scores based in keyword
        """
        score = defaultdict(float)

        for key in IterModel._term_occurrences.keys():
            pir = Probabilistic.catch_pir(vocabulary, key)
            qir = Probabilistic.catch_qir(vocabulary, key)

            score[key] = log(pir/(1-pir), 10) + log((1 - qir)/qir, 10)

            # print("\t\tKey {0} => pir : {1} , qir {2} , score : {3}".format(key, pir, qir, score[key]))

        return score

    @staticmethod
    def __stop() -> bool:
        """
            Stop condition of model
        :return: boolean flag indicates if the model needs to stop
        """
        for key in IterModel._pointers.keys():

            # print("key : {0}, pointer : {1}, occurrences : {2}".format(key, IterModel._pointers[key],
            #                                                            IterModel._term_occurrences[key]))
            if IterModel._pointers[key] < len(IterModel._term_occurrences[key]):
                # print("========================================================")
                return False
        # print("========================================================")
        return True

    @staticmethod
    def __atualize(doc: str, term_scores: DefaultDict[str, float]) -> float:
        """
            Atualize a lowest document score based in term scores
        :param doc: document to atualize
        :param term_scores: defaultdict of term scores
        :return: score of this document
        """
        score = 0.0
        for key in IterModel._term_occurrences.keys():
            try:
                occ = IterModel._term_occurrences[key][IterModel._pointers[key]]
                if occ.doc() == doc:
                    score += term_scores[key]
                    IterModel._pointers[key] += 1
            except ValueError:
                continue
            except IndexError:
                continue
        return score

    @staticmethod
    def __iter_rank(vocabulary: Vocabulary) -> List[Result]:
        """
            One iteration of Probabilistic model execute. Instable in first versions.
        :param vocabulary: vocabulary structure
        :return: ranked list of documents, scores
        """
        term_scores = Probabilistic.__generate_term_scores(vocabulary)

        scores = defaultdict(float)
        while not Probabilistic.__stop():
            doc = IterModel.lowest_doc()

            # print("lwst : {0}".format(doc))

            scores[doc] = Probabilistic.__atualize(doc, term_scores)

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)

        return IterModel.cast_solution(scores)

    @staticmethod
    def __take(rank: List[Result], n: int) -> List[Result]:
        """
            Take the first n elements of this list copying to another structure
        :param rank: list of document, score
        :param n: number of elements to take
        :return: sliced list
        """
        cont = 0
        copy = list()
        for i in rank:
            if cont == n:
                return copy
            cont += 1
            copy.append(i)
        return copy

    @staticmethod
    def catch_pir(vocabulary: Vocabulary, key: str) -> float:
        """
            Catch pir statistic based in key of vocabulary
        :param vocabulary: data structure that represents vocabulary
        :param key: keyword to catch pir
        :return: float pir
        """
        n_value = len(vocabulary.file_names)
        ni = len(Probabilistic._term_occurrences[key])

        if Probabilistic._V:
            vi_value = Probabilistic.docs_with_key(vocabulary[key])
            v_value = len(Probabilistic._V)
            # print("Key {0} => V {1} Vi {2}".format(key, v_value, vi_value))
            return (vi_value + (ni / n_value)) / (v_value + 1)
        else:
            return 0.5

    @staticmethod
    def catch_qir(vocabulary: Vocabulary, key: str) -> float:
        """
            Catch qir statistic based in key of vocabulary
        :param vocabulary: data structure that represents vocabulary
        :param key: keyword to catch pir
        :return: float pir
        """
        n_value = len(vocabulary.file_names)
        ni = len(Probabilistic._term_occurrences[key])

        if Probabilistic._V:
            vi_value = Probabilistic.docs_with_key(vocabulary[key])
            v_value = len(Probabilistic._V)
            return (ni - vi_value + (ni / v_value)) / (n_value - v_value + 1)
        else:
            return ni / n_value

    @staticmethod
    def docs_with_key(occ) -> int:
        """
            Return the vi_value : number of documents with key
        :param occ: Occurrences
        :return:
        """
        doc = 0
        # print("Occurrences : {0}".format(occ))
        for oc in occ:
            if oc.doc() in [t.document for t in Probabilistic._V]:
                doc += 1
        return doc



"""
    First abstraction that represents IR models
        Model, IterModel
"""

import abc

from enum import Enum
from typing import List
from collections import defaultdict

from matchup.structure.Solution import Result
from matchup.structure.Vocabulary import Vocabulary
from matchup.presentation.Text import Term


class ModelType(Enum):
    Boolean = 1,
    Vector = 2,
    Probabilistic = 3,
    ExtendedBoolean = 4

    @classmethod
    def get_type(cls, value: str) -> "ModelType":
        if value.lower() == 'boolean':
            return cls.Boolean
        elif value.lower() == 'vector':
            return cls.Vector
        elif value.lower() == 'probabilistic':
            return cls.Probabilistic
        elif value.lower() == 'extendedboolean':
            return cls.ExtendedBoolean


class NoSuchModelException(RuntimeError):
    pass


class ModelExecutionError(RuntimeError):
    pass


class Model(abc.ABC):

    @staticmethod
    @abc.abstractmethod
    def run(query: List[Term], vocabulary: Vocabulary):
        """
            Define the principal method of IR models.
        :param query: List of all entry terms
        :param vocabulary: Vocabulary pre-processed
        :return:
        """
        ...

    @staticmethod
    def cast_solution(structure: List[tuple]) -> List[Result]:
        return [Result(item[0], round(item[1], 3)) for item in structure]


"""
    Describe one variation of Model classes : IterModel classes have some funcionalities for help his works
    Pointers and occurrences are implemented here.
"""


class IterModel(Model):

    _term_occurrences = defaultdict(list)
    _pointers = defaultdict(int)

    @staticmethod
    @abc.abstractmethod
    def run(query: List[Term], vocabulary: Vocabulary):
        """
            Define the principal method of IR models.
        :param query: List of all entry terms
        :param vocabulary: Vocabulary pre-processed
        :return:
        """
        ...

    @staticmethod
    def initialize_occurrences(query: List[Term], vocabulary: Vocabulary) -> None:
        """
            Create another data structure _term_occurrences that represents the vocabulary wiith just query
            keywords.
        :param query: original query
        :param vocabulary: original vocabulary
        :return: None
        """
        for key in query:
            if key.word in vocabulary:
                IterModel._term_occurrences[key.word] = vocabulary[key.word]

    @staticmethod
    def initialize_pointers() -> None:
        """
            Initialize pointers to model algorithm
        :return: None
        """
        for query_term in IterModel._term_occurrences.keys():
            IterModel._pointers[query_term] = 0

    @classmethod
    def lowest_doc(cls) -> str:
        """
            Return the lowest doc pointer by pointers
        :return: lowest document
        """
        # print("candidates : ", end='')
        lowest = None
        for key in cls._term_occurrences.keys():
            try:
                doc = cls._term_occurrences[key][cls._pointers[key]].doc()
                # print(doc + " ", end='')
                if lowest and doc < lowest:
                    lowest = doc
                elif not lowest:
                    lowest = doc
            except ValueError:
                continue
            except IndexError:
                continue
        # print(".")
        return lowest



from math import pow

from typing import List
from collections import defaultdict

from matchup.presentation.Text import Term
from matchup.models.Model import IterModel
from matchup.structure.Vocabulary import Vocabulary

from matchup.models.algorithms.Vector import Vector


class ExtendedBoolean(IterModel):

    P = 0.0

    @staticmethod
    def run(query: List[Term], vocabulary: Vocabulary):
        IterModel.initialize_occurrences(query, vocabulary)
        IterModel.initialize_pointers()

        IterModel._term_occurrences = Vector.process_documents(query, vocabulary)

        scores = defaultdict(float)

        while not Vector.stop():
            doc = IterModel.lowest_doc()
            doc_repr = Vector.process_doc(doc)
            scores[doc] = ExtendedBoolean.generate_scores(doc_repr, len(query))

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)
        return IterModel.cast_solution(scores)

    @classmethod
    def generate_scores(cls, doc_repr, m) -> float:
        score = 0.0
        for key in doc_repr:
            score += pow((1 - doc_repr[key]), cls.P)

        compensation = m - len(doc_repr)
        if compensation:
            score += compensation

        score /= m
        score = pow(score, 1/cls.P)
        return 1.0 - score









"""
    Describes the data structure of IR models design.
"""
from os import path, listdir
from collections import defaultdict
from typing import List

import pickle

from matchup.structure.Occurrence import Occurrence
from matchup.structure.weighting.Idf import IDF
from matchup.structure.weighting.Tf import TF
from matchup.presentation.Sanitizer import Sanitizer
from matchup.presentation.Text import Term
from matchup.structure.File import ExceptionNotSupported, File

LIB_PATH = path.abspath("./static/lib")


class Vocabulary:

    def __init__(self, *, settings_path: str = None, processed_path: str = None, language: str = None):
        """
        :param settings_path: Path with configurations files like stopwords txt file.
        :param processed_path: Path with folder where the processed vocabulary will save it structure
        :param language: Language of stopwords to process vocabulary
        """
        self._inverted_file = defaultdict(list)
        self._idfs = None
        self._tfs = None

        self.max_frequency_map = defaultdict(float)  # Posso excluir tranquilamente, basta apagar em _push

        stopwords_language = language if language is not None else 'pt-br'
        stopwords_path = settings_path if settings_path else LIB_PATH
        self._sanitizer = Sanitizer(stopwords_path=stopwords_path + f"\\{stopwords_language}.txt")

        prefix = processed_path if processed_path else settings_path
        self._inverted_file_path = prefix + "\\inverted.bin"

        self.file_names = set()
        
    def import_file(self, file_path: str) -> bool:
        """
            Given a file path of a document, this function append this document into some structure, case the path are
            correct. The processing of this file can be started running function generate_vocabulary()

        :param file_path: string that represents a relative or absolute path of an txt file
        :return: boolean flag that indicates if the file has been identified
        """
        if path.exists(file_path):
            self.file_names.add(file_path)
            return True
        return False

    def import_folder(self, folder_path: str, *, update_path: bool = False) -> bool:
        """
            Generalization of import_file(). This function receive a folder path and try to append all documents of
            this folder into some structure. he processing of all this file can be started running function
             generate_vocabulary()

        :param folder_path: string that represents a relative or absolute path of an folder
        :param update_path: flag to update processed file path
        :return: boolean flag that indicates if the folder has been identified
        """
        if path.isdir(folder_path):
            list_dir = filter(lambda x: 'inverted.bin' not in x, listdir(folder_path))
            for filename in list_dir:
                self.import_file(folder_path + "\\" + filename)
            if update_path:
                self._inverted_file_path = folder_path + "\\inverted.bin"
            return True
        raise FileNotFoundError

    def import_vocabulary(self) -> bool:
        """
            This is a function that recover the vocabulary previously generated.
        :return: boolean flag that indicates success or failure in case the vocabulary has no generated yet.
        """
        self._inverted_file.clear()
        if path.exists(self._inverted_file_path):
            with open(self._inverted_file_path, mode='rb') as file:
                self._inverted_file = pickle.load(file)
                self._retrieve_file_names()
            return True
        raise FileNotFoundError

    def generate_vocabulary(self) -> None:
        """
            This function try to process all content of files that have been inserted before, generating
            the vocabulary data structure ready for use.
        :return: None
        """
        for file_name in self.file_names:
            file = None
            try:
                file = File.open(file_name)
                content_file = File.content_file(file_name, file)
                self._process_file(file_name, content_file)
            except ExceptionNotSupported:
                continue
            finally:
                File.close(file)

    def save(self) -> bool:
        """
            Persist data structure on disc.
        :return: boolean flag that indicates if the data structure can be persisted.
        """
        self._sort()
        if self._inverted_file:
            with open(self._inverted_file_path, mode='wb') as file:
                pickle.dump(self._inverted_file, file)
            return True
        raise ReferenceError

    def size_vocabulary(self) -> int:
        """
            Get the number of words this data structure almost process
        :return: Size of data structure vocabulary
        """
        return len(self._inverted_file)

    @property
    def stopwords_path(self) -> str:
        """
        :return: Get the absolute path that stop words are persisted.
        """
        return self._sanitizer.stopwords_path

    @stopwords_path.setter
    def stopwords_path(self, stp_path: str) -> None:
        """
        :return: Get the absolute path that stop words are persisted.
        """
        self._sanitizer.stopwords_path = stp_path

    @property
    def idf(self) -> IDF:
        """
            Get the data structure that represents the IDF weithing
        :return: IDF object
        """
        return self._idfs

    @idf.setter
    def idf(self, idf: IDF) -> None:
        """
            This function just calculate IDF of all keywords on vocabulary
        :return: None
        """
        self._idfs = idf
        self._idfs.generate(self)

    @property
    def tf(self) -> TF:
        """
            Get the data structure that represents the TF weithing
        :return: TF object
        """
        return self._tfs

    @tf.setter
    def tf(self, tf: TF) -> None:
        """
            Set the data structure that represents the TF weithing
        :param tf: TF Object
        :return: None
        """
        self._tfs = tf

    @property
    def keys(self) -> list:
        """
            Get all keywords presents in vocabulary
        :return: list of all keywords
        """
        return list(self._inverted_file.keys())

    @property
    def sanitizer(self) -> "Sanitizer":
        return self._sanitizer

    def get_number_docs_by_keyword(self, kw: str) -> int:
        """
            Get the number of documents contains some keyword
        :param kw: string that represents some keyword
        :return: number of documents
        """
        return len(self._inverted_file[kw])

    def __str__(self) -> str:
        """
            Transform an Vocabulary object in an String
        :return: String that represents the structure of vocabulary
        """
        vocabulary = ""
        for key in self._inverted_file.keys():
            vocabulary += str(key + ":" + str(self._inverted_file[key]) + "\n")
        return vocabulary

    def __contains__(self, item: str) -> bool:
        """
            This function enables the user to make the associative operation with 'in'
        :param item: keyword
        :return: boolean flag that return true if the keyword are in vocabulary
        """
        return item in self._inverted_file

    def __getitem__(self, item: str) -> List[Occurrence]:
        """
            Get some vocabulary occurrences by item, or instantiante it on data structure
        :param item: keyword
        :return: Occurrences of keyword
        """
        return self._inverted_file[item]

    def _push(self, terms: List[Term], file_name: str) -> None:
        """
            Auxiliar function that push all file terms in vocabulary
        :param terms: list with all file terms
        :param file_name: path of file
        :return: None
        """
        for term in terms:
            try:
                idx = self._inverted_file[term.word].index(file_name)
                occurrence = self._inverted_file[term.word][idx]
                occurrence.add(position=term.position)

                if self.max_frequency_map[file_name] < occurrence.frequency:
                    self.max_frequency_map[file_name] = occurrence.frequency

            except ValueError:
                occurrence = Occurrence(file_name, term)
                self._inverted_file[term.word].append(occurrence)

                if self.max_frequency_map[file_name] < occurrence.frequency:
                    self.max_frequency_map[file_name] = occurrence.frequency

    def _sort(self) -> None:
        """
            Order the documents in vocabulary structure.
        :return: None
        """
        for key in self._inverted_file:
            self._inverted_file[key] = sorted(self._inverted_file[key], key=Occurrence.doc)

    def _retrieve_file_names(self) -> None:
        """
            Iterate the data structure retrieving the file names that generated it vocabulary
        :return: None
        """
        for keyword in self._inverted_file:
            for occurrence in self._inverted_file[keyword]:
                self.file_names.add(occurrence.doc())
                if self.max_frequency_map[occurrence.doc()] < occurrence.frequency:
                    self.max_frequency_map[occurrence.doc()] = occurrence.frequency

    def _process_file(self, filename, content_file):
        number_line = 1
        for content_line in content_file:
            terms = self._sanitizer.sanitize_line(content_line, number_line)
            self._push(terms, filename)
            number_line += 1



"""
    Classic IR model. Vectorial model are implemented here, based in Interface Model
"""
from math import sqrt

from collections import defaultdict
from typing import List, DefaultDict

from matchup.models.Model import IterModel, Term, Vocabulary
from matchup.structure.Solution import Result


class Vector(IterModel):

    @staticmethod
    def run(query: List[Term], vocabulary: Vocabulary) -> List[Result]:
        """
           Principal method that represents IR vectorial model.
        :param query: list of all query terms
        :param vocabulary: data structure that represents the vocabulary
        :return: list of solution -> (document, score)
        """
        idfs = vocabulary.idf

        query_repr = Vector.process_query(query, idfs)  # processa a consulta : gera o vetor da consulta

        IterModel.initialize_occurrences(query, vocabulary)
        IterModel.initialize_pointers()

        IterModel._term_occurrences = Vector.process_documents(query, vocabulary)

        scores = defaultdict(float)

        while not Vector.stop():

            doc = IterModel.lowest_doc()
            doc_repr = Vector.process_doc(doc)

            scores[doc] = Vector.generate_scores(doc_repr, query_repr)

        scores = sorted(scores.items(), key=lambda v: v[1], reverse=True)
        return IterModel.cast_solution(scores)

    @staticmethod
    def process_query(query: List[Term], idfs) -> DefaultDict[str, float]:
        """
            Construct query representation
        :param query: list of all terms
        :param idfs: structure IDF
        :return: query representation
        """
        query_repr = defaultdict(float)
        for key in query:
            query_repr[key.word] = idfs[key.word]
        return query_repr

    @staticmethod
    def process_doc(chosen_doc: str) -> DefaultDict[str, float]:
        """
            Proccess doc generating it score
        :param chosen_doc: Str represents the lowest document
        :return: doc repr. Dictionary with all term scores by doc. That is the document vector.
        """
        doc_repr = defaultdict(float)
        for key in IterModel._term_occurrences.keys():
            try:
                occ = IterModel._term_occurrences[key][IterModel._pointers[key]]
                if occ.doc() == chosen_doc:
                    IterModel._pointers[key] += 1
                    doc_repr[key] = occ.score
            except ValueError:
                continue
            except IndexError:
                continue
        return doc_repr

    @staticmethod
    def generate_scores(doc_repr: DefaultDict[str, float], query_repr: DefaultDict[str, float]) -> float:
        """
            Calculate the similarity between one doc and one query by its representations
        :param doc_repr: doc repr
        :param query_repr: query repr
        :return: similarity
        """
        norm_query = 0
        norm_doc = 0
        intern_product = 0

        for key in query_repr.keys():
            intern_product += (doc_repr[key] * query_repr[key])
            norm_doc += doc_repr[key] ** 2
            norm_query += query_repr[key] ** 2

        norm_query = sqrt(norm_query)
        norm_doc = sqrt(norm_doc)

        if norm_doc and norm_query:
            return intern_product / (norm_doc * norm_query)
        else:
            return 0.0

    @staticmethod
    def stop() -> bool:
        """
            Given a dictionary with all pointers by keyword and another dictionary with all occurrences by keyword,
            this function calculate if the algortihm is over
        :return: boolean flag indicates if algorithm stops
        """
        for key in IterModel._pointers.keys():

            # print("key : {0}, pointer : {1}, occurrences : {2}".format(key, IterModel._pointers[key],
            #                                                            IterModel._term_occurrences[key]))
            if IterModel._pointers[key] < len(IterModel._term_occurrences[key]):
                # print("========================================================")
                return False
        # print("========================================================")
        return True

    @staticmethod
    def process_documents(query: List[Term], vocabulary: Vocabulary) -> DefaultDict[str, list]:
        """
            Generate document scores based in query
        :param query: query representation
        :param vocabulary: vocabulary structure
        :return: List of occurrences
        """
        target = defaultdict(list)

        idf = vocabulary.idf
        tf = vocabulary.tf

        for key in query:
            if key.word in vocabulary:
                occurrences = vocabulary[key.word]
                for occurrence in occurrences:
                    score = idf[key.word] * tf.calculate(key.word, occurrence, vocabulary)  # occurrence.frequency=fij
                    occurrence.score = score
                target[key.word] = occurrences
        return target



"""
    Represents the Query of the IR service, composite for :
        sanitizer, orquestrator and answer
"""

from matchup.presentation.Sanitizer import Sanitizer
from matchup.models.Orchestrator import Orchestrator, ModelType
from matchup.structure.Solution import Solution


class NoSuchAnswerException(RuntimeError):
    pass


class Query:
    def __init__(self, *, vocabulary):
        self._sanitizer = Sanitizer(stopwords_path=vocabulary.stopwords_path)
        self._orquestrator = Orchestrator(vocabulary)
        self._answer = list()  # List[Term]

    def ask(self, answer: str = None) -> None:
        """
            Make query since a command line prompt.
        :return: None
        """
        if not answer:
            message = "{0}\n{1: >18}".format(25*"= ", "Consulta: ")
            answer = input(message)
            self._answer = self._sanitizer.sanitize_line(answer, 1)
        else:
            number_line = 1
            text = answer.split("\n")
            terms = []
            for line in text:
                terms += self._sanitizer.sanitize_line(line, number_line)  # linha -> [(palavra, posicao)]
                number_line += 1
            self._answer = terms

        self._orquestrator.entry = self._answer

    @property
    def search_input(self):
        return self._answer

    def search(self, *, model: ModelType = None, idf=None, tf=None, **kwargs) -> Solution:
        """
            Receive an IR model and execute the query based in user answer and the vocabulary.
        :param model: ModelType that represents the IR model
        :param idf: Describe the class  of IDF
        :param tf: Describe the class of TF
        :return: list of solution -> (document, score)
        """
        results = self._orquestrator.search(model, idf, tf, **kwargs)
        return Solution(results)



from abc import ABC
from math import log
from collections import defaultdict


class IDFFactory:
    @staticmethod
    def create_idf_by_str(value: str) -> "IDF":
        if value.lower() == 'unary':
            return Unary()
        elif value.lower() == 'inversefrequency':
            return InverseFrequency()
        elif value.lower() == 'inversefrequencysmooth':
            return InverseFrequencySmooth()
        elif value.lower() == 'inversefrequencymax':
            return InverseFrequencyMax()
        elif value.lower() == 'probabilisticinversefrequency':
            return ProbabilisticInverseFrequency()


class IDF(ABC):

    def __init__(self):
        self._structure = defaultdict(float)

    def __repr__(self):
        string = ""
        for key in self._structure:
            string += f"{key} : {self._structure[key]}\n"
        return string

    def generate(self, vocabulary) -> None:
        """
            Generate the idf for all keywords in a vocabulary
        :param vocabulary: Vocabulary object
        :return: None
        """
        ...

    def take(self, *, value: int = 0, reverse: bool = True):
        """
            Take the first 'value' elements of IDF dict
        :param value:
        :param reverse : asc or desc.
        :return:
        """
        if value != 0:
            return sorted(self._structure.items(), key=lambda v: v[1], reverse=reverse)[:value]
        return sorted(self._structure.items(), key=lambda v: v[1], reverse=reverse)

    def __getitem__(self, item: str) -> float:
        """
            Overload operator [] on idf structure
        :param item: keyword (document)
        :return: float idf value
        """
        return self._structure[item]


class Unary(IDF):

    def __init__(self):
        super(Unary, self).__init__()

    def generate(self, vocabulary):
        """
            Model to calculate IDF based in unary weight (1)
        :param vocabulary: structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._structure[key] = 1


class InverseFrequency(IDF):

    def __init__(self):
        super(InverseFrequency, self).__init__()

    def generate(self, vocabulary):
        """
             Model to calculate IDF based in inverse frequency : log N / ni
        :param vocabulary:  structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._structure[key] = log(len(vocabulary.file_names) / vocabulary.get_number_docs_by_keyword(key), 10)


class InverseFrequencySmooth(IDF):

    def __init__(self):
        super(InverseFrequencySmooth, self).__init__()

    def generate(self, vocabulary):
        """
            Model to calculate IDF based in inverse frequency smooth : log 1 + N / ni
        :param vocabulary: structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._structure[key] = \
                log(1 + (len(vocabulary.file_names) / vocabulary.get_number_docs_by_keyword(key)), 10)


class InverseFrequencyMax(IDF):

    def __init__(self):
        super(InverseFrequencyMax, self).__init__()

    def generate(self, vocabulary):
        """
                Model to calculate IDF based in max inverse frequency  : log 1 + max ni / ni
        :param vocabulary:  structure to generate IDF
        :return: None
        """
        all_number_docs_by_keyword = [vocabulary.get_number_docs_by_keyword(key) for key in vocabulary.keys]
        max_ni = max(all_number_docs_by_keyword)

        for key in vocabulary.keys:
            self._structure[key] = \
                log(1 + (max_ni / vocabulary.get_number_docs_by_keyword(key)), 10)


class ProbabilisticInverseFrequency(IDF):

    def __init__(self):
        super(ProbabilisticInverseFrequency, self).__init__()

    def generate(self, vocabulary) -> None:
        """
            Model to calculate IDF based in probabilistic inverse frequency : log N - ni / ni
        :param vocabulary: structure to generate IDF
        :return: None
        """
        for key in vocabulary.keys:
            self._structure[key] = \
                log((len(vocabulary.file_names) - vocabulary.get_number_docs_by_keyword(key))
                    / vocabulary.get_number_docs_by_keyword(key), 10)




import os
import functools
from typing import List

import PyPDF2


SUPPORTED_EXTENSIONS = ['.txt', '.pdf']


class ExceptionNotSupported(RuntimeError):
    ...


class File:

    @staticmethod
    def get_extension(filename: str) -> str:
        extension = -1
        return os.path.splitext(filename)[extension]

    @staticmethod
    def open(filename: str):
        """
        Define the mode to open the file based in its extension.
        :param filename:
        :return: 'r' or 'rb'
        """
        extension = File.get_extension(filename)
        if extension not in SUPPORTED_EXTENSIONS:
            raise ExceptionNotSupported(f"{extension} is not supported.")

        return open(filename, mode='r', encoding='utf-8') if extension == '.txt' else open(filename, mode='rb')

    @staticmethod
    def close(file):
        if file:
            file.close()

    @staticmethod
    def content_file(filename: str, file):
        extension = File.get_extension(filename)
        if extension not in SUPPORTED_EXTENSIONS:
            raise ExceptionNotSupported(f"{extension} is not supported.")

        if extension == '.txt':
            return file

        pdf_file_reader = PyPDF2.PdfFileReader(file)
        return PDF(pdf_file_reader).all()


class PDF:
    """TODO : reader pagination of .txt and .pdf files"""
    def __init__(self, file_reader):
        self._reader = file_reader
        self._page = 0

    def all(self):
        text = []
        for i in range(self._reader.numPages):
            text += self._reader.getPage(i).extractText().split("\n")
        return text





import unittest
import os

from matchup.structure.Solution import Result
from matchup.structure.Vocabulary import Vocabulary
from matchup.structure.Query import Query
from matchup.models.Model import ModelType
from matchup.structure.weighting.Tf import TermFrequency
from matchup.structure.weighting.Idf import InverseFrequency


class VectorTest(unittest.TestCase):
    def setUp(self):
        self._vocabulary = Vocabulary(settings_path=os.path.abspath("tests/static"),
                                      processed_path=os.path.abspath("tests/static/files"))
        self._vocabulary.import_vocabulary()
        self._query = Query(vocabulary=self._vocabulary)

    def test_search_known_response(self):
        self._query.ask(answer="artilheiro brasil 1994 gols")
        response = self._query.search(model=ModelType.Vector, idf=InverseFrequency(), tf=TermFrequency())

        some_expected_results = [Result(os.path.abspath("tests/static/files/d1.txt"), 1.0),
                                 Result(os.path.abspath("tests/static/files/d3.txt"), 0.808),
                                 Result(os.path.abspath("tests/static/files/d15.txt"), 0.74),
                                 Result(os.path.abspath("tests/static/files/d11.txt"), 0.604)]

        for expected in some_expected_results:
            self.assertTrue(expected in response)



"""
    The brain of IR algorithms. This module are resposible to execute one model and return the resulted
    scored document list.
"""
from typing import List

from matchup.structure.Solution import Result
from matchup.structure.weighting.Idf import InverseFrequency
from matchup.structure.weighting.Tf import LogNormalization

from matchup.models.Model import ModelType
from matchup.models.algorithms.Boolean import Boolean
from matchup.models.algorithms.Vector import Vector
from matchup.models.algorithms.Probabilistic import Probabilistic
from matchup.models.algorithms.ExtendedBoolean import ExtendedBoolean

from matchup.presentation.Text import Term


class NoSuchInputException(RuntimeError):
    pass


class NoSuchModelException(RuntimeError):
    pass


class ModelMissingParameters(RuntimeError):
    pass


class Orchestrator:

    def __init__(self, vocabulary):
        self._vocabulary = vocabulary
        self._input = List[Term]

    def search(self, model: ModelType = None, idf=None, tf=None, **kwargs) -> List[Result]:
        """
            Core function. Execute one IR model based in vocabulary and input(query)
        :param model: IR model to execute
        :param idf: IDF class to weighting IR model
        :param tf: TF class to weighting IR model
        :return: list of solution -> (document, score)
        """

        # setting algorithms IDF weighting
        self._vocabulary.tf = tf if tf is not None else LogNormalization()
        self._vocabulary.idf = idf if idf is not None else InverseFrequency()

        model = model if model else ModelType.Vector

        if self._input:
            if model == ModelType.Boolean:
                return self.run_boolean()
            elif model == ModelType.Vector:
                return self.run_vector_space()
            elif model == ModelType.Probabilistic:
                return self.run_probabilistic()
            elif model == ModelType.ExtendedBoolean:
                return self.run_extended_boolean(**kwargs)
            else:
                raise NoSuchModelException("No model algorithm found. Try again!")
        else:
            raise NoSuchInputException("You should to put some search. Try again!")

    @property
    def entry(self) -> List[Term]:
        """
            Property getter entry(user input)
        :return: user input
        """
        return self._input

    @entry.setter
    def entry(self, etr: List[Term]) -> None:
        """
            Setter attribute input.
        :param etr: input terms
        :return: None
        """
        self._input = etr

    def run_boolean(self):
        return Boolean.run(self._input, self._vocabulary)

    def run_vector_space(self):
        return Vector.run(self._input, self._vocabulary)

    def run_probabilistic(self):
        return Probabilistic.run(self._input, self._vocabulary)

    def run_extended_boolean(self, **kwargs):
        P = kwargs.get("P")
        if P:
            ExtendedBoolean.P = float(P)
            return ExtendedBoolean.run(self._input, self._vocabulary)
        else:
            raise ModelMissingParameters("Missing parameter P for Extended Boolean Model")


"""
    Module responsible to configure the document representation.
"""

import re
import unicodedata

from typing import List, Set

from matchup.presentation.Text import Term, Line


class Sanitizer:
    def __init__(self, *, stopwords_path: str):
        self._stopwords_path = stopwords_path
        self._stopwords = self.import_stopwords()

    def import_stopwords(self) -> Set[str]:
        """
            Retrieve stopwords from memory
        :return: set of stopwords
        """
        with open(self._stopwords_path, encoding='utf-8') as file:
            self._stopwords = {line.strip() for line in file}
        return self._stopwords

    @staticmethod
    def _remove_special_chars_lower(word: str) -> str:
        """
            Given an word, this function clean this string, removing the special chars
        :param word: string to be sanitized
        :return:
        """
        return re.sub('[^a-zA-Z0-9\n]', '', word).lower()

    @staticmethod
    def index_line(words: List[str], base_line: Line) -> List[Term]:
        """
            This function index one line, clean all words and returning all words sanitized
            The list must be sorted by occurrence in text!
        :param words: base-line stripped and without stopwords
        :param base_line: line to be sanitized
        :return: list of indexed words : list(Term)
        """

        # importante para o find funcionar
        base_line_stripped = Sanitizer.strip_accents(base_line.content)

        indexed_words = []

        acc_value = 0
        for word in words:

            word_position = base_line_stripped[acc_value::].find(word)
            acc_value += word_position

            position = str(base_line.number) + "-" + str(acc_value)

            word_sanitized = Sanitizer._remove_special_chars_lower(word)

            if word_sanitized:
                # salvo a palavra minúscula e a posição de ocorrência dentro da linha
                indexed_words.append(Term(word_sanitized, position))

        return indexed_words

    def sanitize_line(self, line: str, number_line: int) -> List[Term]:
        """
            This function sanitize one line. The number is basically for presentation
        :param line: Complete line to be processed
        :param number_line: number of line
        :return:
        """
        base_line = Line(line, number_line)

        line_cleaned = base_line.content.strip()
        line_cleaned = self.strip_accents(line_cleaned)

        words = line_cleaned.split()
        filtered = filter(lambda word: word.lower() not in self._stopwords, words)  # remove stopwords
        # dado a lista de palavras filtradas e a linha original : retorna um indexador de termos
        indexed = Sanitizer.index_line(list(filtered), base_line)
        return indexed  # lista de termos

    @property
    def stopwords_path(self) -> str:
        """
            Property that get the stopwords file path
        :return: Complete stopwords file path
        """
        return self._stopwords_path

    @stopwords_path.setter
    def stopwords_path(self, stopwordspath: str) -> None:
        """
            Property setter stopwords path
        :param stopwordspath: New stopwords path
        :return: None
        """
        self._stopwords_path = stopwordspath

    @staticmethod
    def strip_accents(text: str) -> str:
        """
            Strip accents of one text
        :param text: old text
        :return: text sanitized
        """
        text = unicodedata.normalize('NFD', text).encode('ascii', 'ignore').decode("utf-8")
        return str(text)

